// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/18AI32SSC1M/driver/ioctl.c $
// $Rev: 53598 $
// $Date: 2023-08-23 16:05:50 -0500 (Wed, 23 Aug 2023) $

// 18AI32SSC1M: Device Driver: source file

#include "main.h"



// macros *********************************************************************

#define	AUTOCAL_START			0x2000
#define	AUTOCAL_PASS			0x4000
#define	INIT_START				0x8000

#define	IRQ0_INIT_DONE			0x0000
#define	IRQ0_AUTOCAL_DONE		0x0001
#define	IRQ0_ACTIVE				0x0008
#define	IRQ0_MASK				0x0007



//*****************************************************************************
static void _icr_mod(dev_data_t* dev, u32 value, u32 mask)
{
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_icr_32, value, mask);
}



//*****************************************************************************
static int _query(dev_data_t* dev, void* arg)
{
	int		ret	= 0;
	s32*	ptr	= (s32*) arg;

	switch (ptr[0])
	{
		default:	ptr[0]	= AI32SSC1M_IOCTL_QUERY_ERROR;
					ret		= -EINVAL;
					break;

		case AI32SSC1M_QUERY_AUTOCAL_BG:	ptr[0]	= dev->cache.autocal_bg;			break;
		case AI32SSC1M_QUERY_AUTOCAL_MS:	ptr[0]	= dev->cache.autocal_ms;			break;
		case AI32SSC1M_QUERY_CHANNEL_MAX:	ptr[0]	= dev->cache.channels_max;			break;
		case AI32SSC1M_QUERY_CHANNEL_QTY:	ptr[0]	= dev->cache.channel_qty;			break;
		case AI32SSC1M_QUERY_COUNT:			ptr[0]	= AI32SSC1M_IOCTL_QUERY_LAST + 1;	break;
		case AI32SSC1M_QUERY_DEVICE_TYPE:	ptr[0]	= dev->board_type;					break;
		case AI32SSC1M_QUERY_DATA_WIDTH:	ptr[0]	= dev->cache.data_width;			break;
		case AI32SSC1M_QUERY_EXT_CLK_DIV:	ptr[0]	= dev->cache.ext_clk_div;			break;
		case AI32SSC1M_QUERY_FGEN_MAX:		ptr[0]	= dev->cache.rate_gen_fgen_max;		break;
		case AI32SSC1M_QUERY_FGEN_MIN:		ptr[0]	= dev->cache.rate_gen_fgen_min;		break;
		case AI32SSC1M_QUERY_FIFO_SIZE:		ptr[0]	= dev->cache.fifo_size;				break;
		case AI32SSC1M_QUERY_FSAMP_MAX:		ptr[0]	= dev->cache.fsamp_max;				break;
		case AI32SSC1M_QUERY_FSAMP_MIN:		ptr[0]	= dev->cache.fsamp_min;				break;
		case AI32SSC1M_QUERY_INIT_MS:		ptr[0]	= dev->cache.initialize_ms;			break;
		case AI32SSC1M_QUERY_MASTER_CLOCK:	ptr[0]	= dev->cache.master_clock;			break;
		case AI32SSC1M_QUERY_NRATE_MASK:	ptr[0]	= dev->cache.rate_gen_nrate_mask;	break;
		case AI32SSC1M_QUERY_NRATE_MAX:		ptr[0]	= dev->cache.rate_gen_nrate_max;	break;
		case AI32SSC1M_QUERY_NRATE_MIN:		ptr[0]	= dev->cache.rate_gen_nrate_min;	break;
		case AI32SSC1M_QUERY_PRETRIG:		ptr[0]	= dev->cache.pretrigger;			break;
		case AI32SSC1M_QUERY_RATE_GEN_QTY:	ptr[0]	= dev->cache.rate_gen_qty;			break;
		case AI32SSC1M_QUERY_V_RANGE:		ptr[0]	= dev->cache.voltage_range;			break;
		case AI32SSC1M_IOCTL_QUERY_LAST:	ptr[0]	= AI32SSC1M_IOCTL_QUERY_LAST;		break;
	}

	return(ret);
}



//*****************************************************************************
static int _init_start(dev_data_t* dev, void* arg)
{
	// Initiate initialization.
	os_reg_mem_tx_u32(NULL, dev->vaddr.gsc_bctlr_32, INIT_START);
	return(0);
}



//*****************************************************************************
int initialize_ioctl(dev_data_t* dev, void* arg)
{
	int			i;
	int			mask;
	long		ms			= dev->cache.initialize_ms + 5000;
	long		ms_total	= ms;
	u32			reg;
	int			ret			= 0;
	os_sem_t	sem;
	int			tmp;
	VADDR_T		va			= dev->vaddr.gsc_bctlr_32;
	int			value;
	gsc_wait_t	wt;

	if ((dev->irq.opened) && (gsc_global.driver_unloading == 0))
	{
		ms_total	*= 2;

		// Safely select the Initialize Done interrupt.
		mask	= IRQ0_MASK | IRQ0_ACTIVE;
		value	= IRQ0_INIT_DONE;
		_icr_mod(dev, value, mask);

		// Wait for the local interrupt.
		os_sem_create(&sem);	// dummy, required for wait operations.
		memset(&wt, 0, sizeof(wt));
		wt.flags		= GSC_WAIT_FLAG_INTERNAL;
		wt.gsc			= AI32SSC1M_WAIT_GSC_INIT_DONE;
		wt.timeout_ms	= ms;
		ret				= gsc_wait_event(dev, &wt, _init_start, NULL, &sem);
		os_sem_destroy(&sem);

		if (wt.flags & GSC_WAIT_FLAG_TIMEOUT)
		{
			ret	= ret ? ret : -ETIMEDOUT;
			printf(	"%s: INITILIZE DONE IRQ TIMED OUT AFTER %ld ms.\n",
					dev->model,
					(long) ms);
		}
	}
	else
	{
		_init_start(dev, NULL);
	}

	// Manually wait for completion as the IRQ wait may have ended early.
	tmp	= gsc_poll_u32(dev, ms, va, INIT_START, 0);

	if (tmp)
	{
		ret	= ret ? ret : tmp;
		printf(	"%s: INITIALIZATION DID NOT COMPLETE WITHIN %ld ms.\n",
				dev->model,
				ms_total);
	}

	// Manually check for completion as the IRQ wait may have ended early.
	reg	= os_reg_mem_rx_u32(dev, va);

	if (reg & INIT_START)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: INITILIZE STILL ACTIVE AFTER %ld ms.\n",
				dev->model,
				ms_total);
	}

	// Initialize the software settings.

	for (i = 0; i < DEV_IO_STREAM_QTY; i++)
	{
		if (dev->io.io_streams[i])
		{
			if (dev->io.io_streams[i]->dev_io_sw_init)
			{
				(dev->io.io_streams[i]->dev_io_sw_init)(dev, dev->io.io_streams[i]);
			}
		}
	}

	return(ret);
}



//*****************************************************************************
static int _autocal_start(dev_data_t* dev, void* arg)
{
	// Initiate autocalibration.
	os_reg_mem_mx_u32(NULL, dev->vaddr.gsc_bctlr_32, AUTOCAL_START, AUTOCAL_START);
	return(0);
}



//*****************************************************************************
static int _autocal(dev_data_t* dev, void* arg)
{
	u32			mask;
	long		ms		= dev->cache.autocal_ms + 5000;
	u32			reg;
	int			ret;
	os_sem_t	sem;
	int			tmp;
	VADDR_T		va		= dev->vaddr.gsc_bctlr_32;
	u32			value;
	gsc_wait_t	wt;

	// Safely select the Autocal Done interrupt.
	mask	= IRQ0_MASK | IRQ0_ACTIVE;
	value	= IRQ0_AUTOCAL_DONE;
	_icr_mod(dev, mask, value);

	// Wait for the local interrupt.
	os_sem_create(&sem);	// dummy, required for wait operations.
	memset(&wt, 0, sizeof(wt));
	wt.flags		= GSC_WAIT_FLAG_INTERNAL;
	wt.gsc			= AI32SSC1M_WAIT_GSC_AUTOCAL_DONE;
	wt.timeout_ms	= ms;
	ret				= gsc_wait_event(dev, &wt, _autocal_start, NULL, &sem);
	os_sem_destroy(&sem);

	if (wt.flags & GSC_WAIT_FLAG_TIMEOUT)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: AUTOCALIBRATE DONE IRQ TIMED OUT AFTER %ld ms.\n",
				dev->model,
				ms);
	}

	// Manually wait for completion in case something terminates our wait early.
	tmp	= gsc_poll_u32(dev, ms, va, AUTOCAL_START, 0);

	if (tmp)
	{
		ms	*= 2;
		ret	= ret ? ret : tmp;
		printf(	"%s: AUTOCALIBRATION DID NOT COMPLETE WITHIN %ld ms.\n",
				dev->model,
				ms);
	}

	// Manually check for completion as the IRQ wait may have ended early.
	reg	= os_reg_mem_rx_u32(dev, va);

	if (reg & AUTOCAL_START)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: AUTOCALIBRATION STILL ACTIVE AFTER %ld ms.\n",
				dev->model,
				ms);
	}

	// Final results.
	reg	= os_reg_mem_rx_u32(dev, va);

	if ((reg & AUTOCAL_PASS) == 0)
	{
		ret	= ret ? ret : -EIO;
		printf(	"%s: AUTOCALIBRATION FAILED (%ld ms).\n",
				dev->model,
				ms);
	}

	if (ret == 0)
	{
		// Wait for settling.
		os_time_sleep_ms(100);
	}

	return(ret);
}



//*****************************************************************************
static int _autocal_bg(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_AUTOCAL_BG_DISABLE,
		AI32SSC1M_AUTOCAL_BG_ENABLE,
		-1	// terminate list
	};

	int	ret;

	if (dev->cache.autocal_bg)
		ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 8, 8);
	else
		ret	= -EINVAL;

	return(ret);
}



//*****************************************************************************
static int _rx_io_mode(dev_data_t* dev, s32* arg)
{
	static const s32	list[]	=
	{
		GSC_IO_MODE_PIO,
		GSC_IO_MODE_BMDMA,
		GSC_IO_MODE_DMDMA,
		-1
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, list, &dev->io.rx.io_mode);
	return(ret);
}



//*****************************************************************************
static int _rx_io_overflow(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_IO_OVERFLOW_IGNORE,
		AI32SSC1M_IO_OVERFLOW_CHECK,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, options, &dev->io.rx.overflow_check);
	return(ret);
}



//*****************************************************************************
static int _rx_io_timeout(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_var(
			arg,
			AI32SSC1M_IO_TIMEOUT_MIN,
			AI32SSC1M_IO_TIMEOUT_INFINITE,
			&dev->io.rx.timeout_s);
	return(ret);
}



//*****************************************************************************
static int _rx_io_underflow(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_IO_UNDERFLOW_IGNORE,
		AI32SSC1M_IO_UNDERFLOW_CHECK,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, options, &dev->io.rx.underflow_check);
	return(ret);
}



//*****************************************************************************
static int _ain_buf_clear(dev_data_t* dev, void* arg)
{
	#define	CLEAR	D18		// IBCR
	#define	OVER	D17		// BCTLR
	#define	UNDER	D16		// BCTLR

	int		i;
	u32		reg;
	int		ret	= 0;
	VADDR_T	va	= dev->vaddr.gsc_ibcr_32;

	// Clear the buffer.
	os_reg_mem_mx_u32(dev, va, CLEAR, CLEAR);

	// Wait for the bit to clear.

	for (i = 0;; i++)
	{
		reg	= os_reg_mem_rx_u32(dev, va);

		if ((reg & CLEAR) == 0)
			break;

		if (i >= 250)
		{
			ret	= -EINVAL;
			printf(	"%s: The analog input buffer took too long to clear.\n",
					DEV_NAME);
			break;
		}
	}

	va	= dev->vaddr.gsc_bctlr_32;

	// Clear the Overflow status bit.
	os_reg_mem_mx_u32(dev, va, 0, OVER);

	// Clear the Underflow status bit.
	os_reg_mem_mx_u32(dev, va, 0, UNDER);

	return(ret);
}



//*****************************************************************************
static int _ain_buf_level(dev_data_t* dev, void* arg)
{
	u32*	ptr	= arg;

	ptr[0]	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bufsr_32);
	ptr[0]	&= 0x7FFFF;
	return(0);
}



//*****************************************************************************
static int _ain_buf_overflow(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_AIN_BUF_OVERFLOW_IGNORE,
		AI32SSC1M_AIN_BUF_OVERFLOW_CLEAR,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 17, 17);
	return(ret);
}



//*****************************************************************************
static int _ain_buf_thr_lvl(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0x3FFFF, dev->vaddr.gsc_ibcr_32, 17, 0);
	return(ret);
}



//*****************************************************************************
static int _ain_buf_thr_sts(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_AIN_BUF_THR_STS_CLEAR,
		AI32SSC1M_AIN_BUF_THR_STS_SET,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ibcr_32, 19, 19);
	return(ret);
}



//*****************************************************************************
static int _ain_buf_underflow(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_AIN_BUF_UNDERFLOW_IGNORE,
		AI32SSC1M_AIN_BUF_UNDERFLOW_CLEAR,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 16, 16);
	return(ret);
}



//*****************************************************************************
static int _ain_mode(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_AIN_MODE_DIFF,
		AI32SSC1M_AIN_MODE_ZERO,
		AI32SSC1M_AIN_MODE_VREF,
		-1	// terminate list
	};

	int	ret;
	s32	v	= arg[0];

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 2, 0);

	if ((v == -1) && (ret == 0))
		os_time_sleep_ms(100);

	return(ret);
}



//*****************************************************************************
static int _ain_range(dev_data_t* dev, s32* arg)
{
	#define	HIGH		0x20
	#define	LOW			0x10
	#undef	MASK
	#define	MASK		0x30

	s32*	ptr	= arg;
	u32		reg;
	int		ret	= 0;
	s32		v	= arg[0];

	switch (v)
	{
		default:

		case -1:

			reg	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bctlr_32);

			switch (reg & MASK)
			{
				default:

					ret	= -EINVAL;
					break;

				case 0x10:

					if (dev->cache.voltage_range == AI32SSC1M_QUERY_V_RANGE_10V)
						ptr[0]	= AI32SSC1M_AIN_RANGE_5V;
					else
						ptr[0]	= AI32SSC1M_AIN_RANGE_1_25V;

					break;

				case 0x20:
				case 0x30:

					if (dev->cache.voltage_range == AI32SSC1M_QUERY_V_RANGE_10V)
						ptr[0]	= AI32SSC1M_AIN_RANGE_10V;
					else
						ptr[0]	= AI32SSC1M_AIN_RANGE_2_5V;

					break;
			}

			break;

		case AI32SSC1M_AIN_RANGE_1_25V:
		case AI32SSC1M_AIN_RANGE_2_5V:

			if (dev->cache.voltage_range == AI32SSC1M_QUERY_V_RANGE_10V)
			{
				ret	= -EINVAL;
				break;
			}

			if (v == AI32SSC1M_AIN_RANGE_2_5V)
				os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bctlr_32, HIGH, MASK);
			else
				os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bctlr_32, LOW, MASK);

			break;

		case AI32SSC1M_AIN_RANGE_5V:
		case AI32SSC1M_AIN_RANGE_10V:

			if (dev->cache.voltage_range == AI32SSC1M_QUERY_V_RANGE_2_5)
			{
				ret	= -EINVAL;
				break;
			}

			if (v == AI32SSC1M_AIN_RANGE_10V)
				os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bctlr_32, HIGH, MASK);
			else
				os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bctlr_32, LOW, MASK);

			break;
	}

	if ((v == -1) && (ret == 0))
		os_time_sleep_ms(100);

	return(ret);
}



//*****************************************************************************
static int _burst_busy(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_BURST_BUSY_IDLE,
		AI32SSC1M_BURST_BUSY_ACTIVE,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_sscr_32, 7, 7);
	return(ret);
}



//*****************************************************************************
static int _burst_size(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0xFFFFF, dev->vaddr.gsc_bstsr_32, 19, 0);
	return(ret);
}



//*****************************************************************************
static int _burst_sync(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_BURST_SYNC_DISABLE,
		AI32SSC1M_BURST_SYNC_RBG,
		AI32SSC1M_BURST_SYNC_EXT,
		AI32SSC1M_BURST_SYNC_BCR,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_sscr_32, 9, 8);
	return(ret);
}



//*****************************************************************************
static int _chan_active(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_CHAN_ACTIVE_SINGLE,
		AI32SSC1M_CHAN_ACTIVE_0_1,
		AI32SSC1M_CHAN_ACTIVE_0_3,
		AI32SSC1M_CHAN_ACTIVE_0_7,
		AI32SSC1M_CHAN_ACTIVE_0_15,
		AI32SSC1M_CHAN_ACTIVE_0_31,
		AI32SSC1M_CHAN_ACTIVE_RANGE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_sscr_32, 2, 0);
	return(ret);
}



//*****************************************************************************
static int _chan_first(dev_data_t* dev, void* arg)
{
	u32	last;
	u32	reg;
	int	ret;

	reg		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_acar_32);
	last	= GSC_FIELD_DECODE(reg, 15, 8);
	ret		= gsc_s32_range_reg(dev, arg, 0, last, dev->vaddr.gsc_acar_32, 7, 0);
	return(ret);
}



//*****************************************************************************
static int _chan_last(dev_data_t* dev, void* arg)
{
	u32	first;
	u32	last;
	u32	reg;
	int	ret;

	last	= dev->cache.channel_qty - 1;
	reg		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_acar_32);
	first	= GSC_FIELD_DECODE(reg, 7, 0);
	ret		= gsc_s32_range_reg(dev, arg, first, last, dev->vaddr.gsc_acar_32, 15, 8);
	return(ret);
}



//*****************************************************************************
static int _chan_single(dev_data_t* dev, void* arg)
{
	u32	last;
	int	ret;

	last	= dev->cache.channel_qty - 1;
	ret		= gsc_s32_range_reg(dev, arg, 0, last, dev->vaddr.gsc_sscr_32, 17, 12);
	return(ret);
}



//*****************************************************************************
static int _clock_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_CLOCK_ENABLE_NO,
		AI32SSC1M_CLOCK_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_sscr_32, 5, 5);
	return(ret);
}



//*****************************************************************************
static int _data_format(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_DATA_FORMAT_2S_COMP,
		AI32SSC1M_DATA_FORMAT_OFF_BIN,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 6, 6);
	return(ret);
}



//*****************************************************************************
static int _data_packing(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_DATA_PACKING_DISABLE,
		AI32SSC1M_DATA_PACKING_ENABLE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 18, 18);
	return(ret);
}



//*****************************************************************************
static int _data_width(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_DATA_WIDTH_16,
		AI32SSC1M_DATA_WIDTH_18,
		-1	// terminate list
	};

	int	ret;

	if (dev->cache.data_width == 18)
	{
		ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ibcr_32, 20, 20);
	}
	else if (arg[0] == -1)
	{
		ret		= 0;
		arg[0]	= AI32SSC1M_DATA_WIDTH_16;
	}
	else if (arg[0] == AI32SSC1M_DATA_WIDTH_16)
	{
		ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ibcr_32, 20, 20);
	}
	else
	{
		ret	= -EINVAL;
	}

	return(ret);
}



//*****************************************************************************
static int _input_sync(dev_data_t* dev, void* arg)
{
	#undef	MASK
	#undef	SHIFT
	#define	MASK	(0x1L << SHIFT)
	#define	SHIFT	12

	unsigned long	ms_limit;
	int				ret;

	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bctlr_32, MASK, MASK);

	// Wait for the operation to complete.

	if (dev->io.rx.timeout_s)
		ms_limit	= dev->io.rx.timeout_s * 1000;
	else
		ms_limit	= 1000;

	ret	= gsc_poll_u32(dev, ms_limit, dev->vaddr.gsc_bctlr_32, MASK, 0);
	return(ret);
}



//*****************************************************************************
static int _io_inv(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_IO_INV_LOW,
		AI32SSC1M_IO_INV_HIGH,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_sscr_32, 11, 11);
	return(ret);
}



//*****************************************************************************
static int _irq0_sel(dev_data_t* dev, s32* arg)
{
	u32	reg;
	int	ret	= 0;

	switch (arg[0])
	{
		default:

			ret	= -EINVAL;
			break;

		case -1:

			// Retrieve the current setting.
			reg		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_icr_32);
			arg[0]	= GSC_FIELD_DECODE(reg, 2, 0);
			break;

		case AI32SSC1M_IRQ0_INIT_DONE:
		case AI32SSC1M_IRQ0_AUTOCAL_DONE:
		case AI32SSC1M_IRQ0_SYNC_START:
		case AI32SSC1M_IRQ0_SYNC_DONE:
		case AI32SSC1M_IRQ0_BURST_START:
		case AI32SSC1M_IRQ0_BURST_DONE:

			// Clear the status bit and apply the new setting.
			// Do not clear the IRQ1 status bit.
			os_reg_mem_mx_u32(dev, dev->vaddr.gsc_icr_32, arg[0] | 0x80, 0x8F);
			break;
	}

	return(ret);
}



//*****************************************************************************
static int _irq1_sel(dev_data_t* dev, s32* arg)
{
	u32	reg;
	int	ret	= 0;

	switch (arg[0])
	{
		default:

			ret	= -EINVAL;
			break;

		case -1:

			// Retrieve the current setting.
			reg		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_icr_32);
			arg[0]	= GSC_FIELD_DECODE(reg, 6, 4);
			break;

		case AI32SSC1M_IRQ1_NONE:
		case AI32SSC1M_IRQ1_IN_BUF_THR_L2H:
		case AI32SSC1M_IRQ1_IN_BUF_THR_H2L:
		case AI32SSC1M_IRQ1_IN_BUF_OVR_UNDR:

			// Clear the status bit and apply the new setting.
			// Do not clear the IRQ0 status bit.
			os_reg_mem_mx_u32(dev, dev->vaddr.gsc_icr_32, (arg[0] << 4) | 0x8, 0xF8);
			break;
	}

	return(ret);
}



//*****************************************************************************
static int _rag_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_GEN_ENABLE_NO,
		AI32SSC1M_GEN_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ragr_32, 16, 16);
	return(ret);
}



//*****************************************************************************
static int _rag_nrate(dev_data_t* dev, s32* arg)
{
	int	ret;
	s32	v	= arg[0];

	ret	= gsc_s32_range_reg(dev, arg, 36, 0xFFFF, dev->vaddr.gsc_ragr_32, 15, 0);

	if ((v == -1) && (ret == 0))
		os_time_sleep_ms(100);

	return(ret);
}



//*****************************************************************************
static int _rbg_clk_src(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_RBG_CLK_SRC_MASTER,
		AI32SSC1M_RBG_CLK_SRC_RAG,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_sscr_32, 10, 10);
	return(ret);
}



//*****************************************************************************
static int _rbg_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_GEN_ENABLE_NO,
		AI32SSC1M_GEN_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_rbgr_32, 16, 16);
	return(ret);
}



//*****************************************************************************
static int _rbg_nrate(dev_data_t* dev, s32* arg)
{
	int	ret;
	s32	v	= arg[0];

	ret	= gsc_s32_range_reg(dev, arg, 36, 0xFFFF, dev->vaddr.gsc_rbgr_32, 15, 0);

	if ((v == -1) && (ret == 0))
		os_time_sleep_ms(100);

	return(ret);
}



//*****************************************************************************
static int _samp_clk_src(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_SAMP_CLK_SRC_EXT,
		AI32SSC1M_SAMP_CLK_SRC_RAG,
		AI32SSC1M_SAMP_CLK_SRC_RBG,
		AI32SSC1M_SAMP_CLK_SRC_BCR,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_sscr_32, 4, 3);
	return(ret);
}



//*****************************************************************************
static int _scan_marker(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_SCAN_MARKER_DISABLE,
		AI32SSC1M_SCAN_MARKER_ENABLE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 11, 11);
	return(ret);
}



//*****************************************************************************
static int _aux_clk_mode(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_AUX_CLK_MODE_DISABLE,
		AI32SSC1M_AUX_CLK_MODE_INPUT,
		AI32SSC1M_AUX_CLK_MODE_OUTPUT,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_asiocr_32, 1, 0);
	return(ret);
}



//*****************************************************************************
static int _aux_in_pol(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_AUX_IN_POL_LO_2_HI,
		AI32SSC1M_AUX_IN_POL_HI_2_LO,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_asiocr_32, 8, 8);
	return(ret);
}



//*****************************************************************************
static int _aux_noise(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_AUX_NOISE_LOW,
		AI32SSC1M_AUX_NOISE_HIGH,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_asiocr_32, 10, 10);
	return(ret);
}



//*****************************************************************************
static int _aux_out_pol(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_AUX_OUT_POL_HI_PULSE,
		AI32SSC1M_AUX_OUT_POL_LOW_PULSE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_asiocr_32, 9, 9);
	return(ret);
}



//*****************************************************************************
static int _aux_sync_mode(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_AUX_SYNC_MODE_DISABLE,
		AI32SSC1M_AUX_SYNC_MODE_INPUT,
		AI32SSC1M_AUX_SYNC_MODE_OUTPUT,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_asiocr_32, 3, 2);
	return(ret);
}



//*****************************************************************************
static int _pretrig(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_PRETRIG_DISABLE,
		AI32SSC1M_PRETRIG_ENABLE,
		-1	// terminate list
	};

	int	ret;

	if (dev->cache.pretrigger)
		ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_sscr_32, 18, 18);
	else
		ret	= -EIO;

	return(ret);
}



//*****************************************************************************
static int _pretrig_latch(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_PRETRIG_LATCH_IDLE,
		AI32SSC1M_PRETRIG_LATCH_ACTIVE,
		-1	// terminate list
	};

	int	ret;

	if (dev->cache.pretrigger)
		ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_sscr_32, 19, 19);
	else
		ret	= -EIO;

	return(ret);
}



//*****************************************************************************
static int _rx_io_abort(dev_data_t* dev, s32* arg)
{
	arg[0]	= gsc_read_abort_active_xfer(dev, &dev->io.rx);
	return(0);
}



//*****************************************************************************
static int _pretrig_count(dev_data_t* dev, ai32ssc1m_pretrig_count_t* arg)
{
	int	ret;

	if (dev->cache.pretrigger)
	{
		ret			= 0;
		arg->low	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_ptclr_32);
		arg->high	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_ptchr_32);
		arg->high	&= 0xFFFF;
	}
	else
	{
		ret	= -EIO;
	}

	return(ret);
}



//*****************************************************************************
static int _scan_marker_get(dev_data_t* dev, u32* arg)
{
	u32	smlwr;
	u32	smuwr;

	smlwr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_smlwr_32);
	smuwr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_smuwr_32);
	arg[0]	= (smuwr << 16) | smlwr;
	return(0);
}



//*****************************************************************************
static int _scan_marker_set(dev_data_t* dev, u32* arg)
{
	u32	smlwr;
	u32	smuwr;

	smlwr	= arg[0] & 0xFFFF;
	smuwr	= arg[0] >> 16;
	os_reg_mem_tx_u32(dev, dev->vaddr.gsc_smlwr_32, smlwr);
	os_reg_mem_tx_u32(dev, dev->vaddr.gsc_smuwr_32, smuwr);
	return(0);
}



//*****************************************************************************
static int _ext_clk_div_ena(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AI32SSC1M_EXT_CLK_DIV_ENA_NO,
		AI32SSC1M_EXT_CLK_DIV_ENA_YES,
		-1	// terminate list
	};

	int	ret;

	if (dev->cache.ext_clk_div)
		ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_sscr_32, 6, 6);
	else
		ret	= -EIO;

	return(ret);
}



//*****************************************************************************
static int _ext_clk_div_val(dev_data_t* dev, void* arg)
{
	int	ret;

	if (dev->cache.ext_clk_div)
		ret	= gsc_s32_range_reg(dev, arg, 0, 0xFFFFF, dev->vaddr.gsc_ecdr_32, 19, 0);
	else
		ret	= -EIO;

	return(ret);
}



//*****************************************************************************
static int _autocal_status(dev_data_t* dev, s32* arg)
{
	u32	reg;

	reg	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bctlr_32);

	if (reg & AUTOCAL_START)
		arg[0]	= AI32SSC1M_AUTOCAL_STATUS_ACTIVE;
	else if (reg & AUTOCAL_PASS)
		arg[0]	= AI32SSC1M_AUTOCAL_STATUS_PASS;
	else
		arg[0]	= AI32SSC1M_AUTOCAL_STATUS_FAIL;

	return(0);
}



// variables ******************************************************************

const gsc_ioctl_t	dev_ioctl_list[]	=
{
	{ AI32SSC1M_IOCTL_REG_READ,				(void*) gsc_reg_read_ioctl		},
	{ AI32SSC1M_IOCTL_REG_WRITE,			(void*) gsc_reg_write_ioctl		},
	{ AI32SSC1M_IOCTL_REG_MOD,				(void*) gsc_reg_mod_ioctl		},
	{ AI32SSC1M_IOCTL_QUERY,				(void*) _query					},
	{ AI32SSC1M_IOCTL_INITIALIZE,			(void*) initialize_ioctl		},
	{ AI32SSC1M_IOCTL_AUTOCAL,				(void*) _autocal				},
	{ AI32SSC1M_IOCTL_AUTOCAL_BG,			(void*) _autocal_bg				},
	{ AI32SSC1M_IOCTL_RX_IO_MODE,			(void*) _rx_io_mode				},
	{ AI32SSC1M_IOCTL_RX_IO_OVERFLOW,		(void*) _rx_io_overflow			},
	{ AI32SSC1M_IOCTL_RX_IO_TIMEOUT,		(void*) _rx_io_timeout			},
	{ AI32SSC1M_IOCTL_RX_IO_UNDERFLOW,		(void*) _rx_io_underflow		},
	{ AI32SSC1M_IOCTL_AIN_BUF_CLEAR,		(void*) _ain_buf_clear			},
	{ AI32SSC1M_IOCTL_AIN_BUF_LEVEL,		(void*) _ain_buf_level			},
	{ AI32SSC1M_IOCTL_AIN_BUF_OVERFLOW,		(void*) _ain_buf_overflow		},
	{ AI32SSC1M_IOCTL_AIN_BUF_THR_LVL,		(void*) _ain_buf_thr_lvl		},
	{ AI32SSC1M_IOCTL_AIN_BUF_THR_STS,		(void*) _ain_buf_thr_sts		},
	{ AI32SSC1M_IOCTL_AIN_BUF_UNDERFLOW,	(void*) _ain_buf_underflow		},
	{ AI32SSC1M_IOCTL_AIN_MODE,				(void*) _ain_mode				},
	{ AI32SSC1M_IOCTL_AIN_RANGE,			(void*) _ain_range				},
	{ AI32SSC1M_IOCTL_BURST_BUSY,			(void*) _burst_busy				},
	{ AI32SSC1M_IOCTL_BURST_SIZE,			(void*) _burst_size				},
	{ AI32SSC1M_IOCTL_BURST_SYNC,			(void*) _burst_sync				},
	{ AI32SSC1M_IOCTL_CHAN_ACTIVE,			(void*) _chan_active			},
	{ AI32SSC1M_IOCTL_CHAN_FIRST,			(void*) _chan_first				},
	{ AI32SSC1M_IOCTL_CHAN_LAST,			(void*) _chan_last				},
	{ AI32SSC1M_IOCTL_CHAN_SINGLE,			(void*) _chan_single			},
	{ AI32SSC1M_IOCTL_CLOCK_ENABLE,			(void*) _clock_enable			},
	{ AI32SSC1M_IOCTL_DATA_FORMAT,			(void*) _data_format			},
	{ AI32SSC1M_IOCTL_DATA_PACKING,			(void*) _data_packing			},
	{ AI32SSC1M_IOCTL_DATA_WIDTH,			(void*) _data_width				},
	{ AI32SSC1M_IOCTL_INPUT_SYNC,			(void*) _input_sync				},
	{ AI32SSC1M_IOCTL_IO_INV,				(void*) _io_inv					},
	{ AI32SSC1M_IOCTL_IRQ0_SEL,				(void*) _irq0_sel				},
	{ AI32SSC1M_IOCTL_IRQ1_SEL,				(void*) _irq1_sel				},
	{ AI32SSC1M_IOCTL_RAG_ENABLE,			(void*) _rag_enable				},
	{ AI32SSC1M_IOCTL_RAG_NRATE,			(void*) _rag_nrate				},
	{ AI32SSC1M_IOCTL_RBG_CLK_SRC,			(void*) _rbg_clk_src			},
	{ AI32SSC1M_IOCTL_RBG_ENABLE,			(void*) _rbg_enable				},
	{ AI32SSC1M_IOCTL_RBG_NRATE,			(void*) _rbg_nrate				},
	{ AI32SSC1M_IOCTL_SAMP_CLK_SRC,			(void*) _samp_clk_src			},
	{ AI32SSC1M_IOCTL_SCAN_MARKER,			(void*) _scan_marker			},
	{ AI32SSC1M_IOCTL_AUX_CLK_MODE,			(void*) _aux_clk_mode			},
	{ AI32SSC1M_IOCTL_AUX_IN_POL,			(void*) _aux_in_pol				},
	{ AI32SSC1M_IOCTL_AUX_NOISE,			(void*) _aux_noise				},
	{ AI32SSC1M_IOCTL_AUX_OUT_POL,			(void*) _aux_out_pol			},
	{ AI32SSC1M_IOCTL_AUX_SYNC_MODE,		(void*) _aux_sync_mode			},
	{ AI32SSC1M_IOCTL_PRETRIG,				(void*) _pretrig				},
	{ AI32SSC1M_IOCTL_PRETRIG_LATCH,		(void*) _pretrig_latch			},
	{ 0,									NULL							},	// no longer supported
	{ AI32SSC1M_IOCTL_RX_IO_ABORT,			(void*) _rx_io_abort			},
	{ AI32SSC1M_IOCTL_VPD_READ,				(void*) gsc_vpd_read_ioctl		},
	{ AI32SSC1M_IOCTL_WAIT_EVENT,			(void*) gsc_wait_event_ioctl	},
	{ AI32SSC1M_IOCTL_WAIT_CANCEL,			(void*) gsc_wait_cancel_ioctl	},
	{ AI32SSC1M_IOCTL_WAIT_STATUS,			(void*) gsc_wait_status_ioctl	},
	{ AI32SSC1M_IOCTL_PRETRIG_COUNT,		(void*) _pretrig_count			},
	{ AI32SSC1M_IOCTL_SCAN_MARKER_GET,		(void*) _scan_marker_get		},
	{ AI32SSC1M_IOCTL_SCAN_MARKER_SET,		(void*) _scan_marker_set		},
	{ AI32SSC1M_IOCTL_EXT_CLK_DIV_ENA,		(void*) _ext_clk_div_ena		},
	{ AI32SSC1M_IOCTL_EXT_CLK_DIV_VAL,		(void*) _ext_clk_div_val		},
	{ AI32SSC1M_IOCTL_AUTOCAL_STATUS,		(void*) _autocal_status			},
	{ -1, NULL }
};


