// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/18AI32SSC1M/samples/sbtest/service.c $
// $Rev: 53596 $
// $Date: 2023-08-23 15:57:48 -0500 (Wed, 23 Aug 2023) $

// 18AI32SSC1M: Sample Application: source file

#include "main.h"



/******************************************************************************
*
*	Function:	_ioctl_get
*
*	Purpose:
*
*		Issue the given IOCTL call to retrieve the current setting, then
*		verify that the setting is as expected.
*
*	Arguments:
*
*		fd		The handle for the device to access.
*
*		line	The line number where the service call is made.
*
*		index	The table index to report, or negactive to ignore.
*
*		cmd		The IOCTL service to request.
*
*		arg		The argument to pass to the service.
*
*	Returned:
*
*		>= 0	The number of errors encounterred.
*
******************************************************************************/

static int _ioctl_get(int fd, int line, int index, unsigned long cmd, s32 arg)
{
	int	errs	= 0;
	int	i;
	s32	get;

	for (;;)	// A convenience loop.
	{
		// Read the current setting.
		get	= -1;
		i	= ai32ssc1m_ioctl(fd, cmd, &get);

		if (i < 0)
		{
			// There was a problem.
			errs	= 1;
			printf("FAIL <--- (%d. ", line);

			if (index >= 0)
				printf("index %d, ", index);

			printf("ai32ssc1m_ioctl() error, status %d)\n", i);
			break;
		}

		if (arg != get)
		{
			// We didn't get the same setting back.
			errs	= 1;
			printf("FAIL <--- (%d. ", line);

			if (index >= 0)
				printf("index %d, ", index);

			printf(	"got %ld (0x%lX), expected %ld (0x%lX))\n",
					(long) get,
					(long) get,
					(long) arg,
					(long) arg);
			break;
		}

		// All went well.
		errs	= 0;
		break;
	}

	return(errs);
}



/******************************************************************************
*
*	Function:	_ioctl_set
*
*	Purpose:
*
*		Issue the given IOCTL call.
*
*	Arguments:
*
*		fd		The handle for the device to access.
*
*		line	The line number where the service call is made.
*
*		index	The table index to report, or negactive to ignore.
*
*		cmd		The IOCTL service to request.
*
*		arg		The argument to pass to the service.
*
*	Returned:
*
*		>= 0	The number of errors encounterred.
*
******************************************************************************/

static int _ioctl_set(int fd, int line, int index, unsigned long cmd, s32 arg)
{
	int	errs	= 0;
	int	i;

	for (;;)	// A convenience loop.
	{
		// Perform the operation.
		i	= ai32ssc1m_ioctl(fd, cmd, &arg);

		if (i == 0)
		{
			// All went well.
			errs	= 0;
			break;
		}

		// There was an error.
		errs	= 1;
		printf("FAIL <--- (%d. ", line);

		if (index >= 0)
			printf("index %d, ", index);

		printf("ai32ssc1m_ioctl(), status %d)\n", i);
		break;
	}

	return(errs);
}



/******************************************************************************
*
*	Function:	_reg_mod
*
*	Purpose:
*
*		Perform a read-modify-write on the specified register.
*
*	Arguments:
*
*		fd		The handle for the device to access.
*
*		line	The line number where the service call is made.
*
*		index	The table index to report, or negactive to ignore.
*
*		reg		The register to access. If -1 then this is ignored.
*
*		mask	The set of register bits of interest.
*
*		value	The value to apply/check for the bits of interest.
*
*	Returned:
*
*		>= 0	The number of errors encounterred.
*
******************************************************************************/

static int _reg_mod(int fd, int line, int index, u32 reg, u32 mask, u32 value)
{
	int			errs	= 0;
	int			i;
	gsc_reg_t	parm;

	for (;;)	// A convenience loop.
	{
		if ((int) reg == -1)
		{
			// We're supposed to ignore this register.
			errs	= 0;
			break;
		}

		// Perform the operation.
		parm.reg	= reg;
		parm.value	= value;;
		parm.mask	= mask;
		i			= ai32ssc1m_ioctl(fd, AI32SSC1M_IOCTL_REG_MOD, &parm);

		if (i == 0)
		{
			// All went well.
			errs	= 0;
			break;
		}

		// There was an error.
		errs	= 1;
		printf("FAIL <--- (%d. ", line);

		if (index >= 0)
			printf("index %d, ", index);

		printf("AI32SSC1M_IOCTL_REG_MOD, status %d)\n", i);
		break;
	}

	return(errs);
}



/******************************************************************************
*
*	Function:	_reg_read
*
*	Purpose:
*
*		Read a register, but do nothing with the value.
*
*	Arguments:
*
*		fd		The handle for the device to access.
*
*		line	The line number where the service call is made.
*
*		index	The table index to report, or negactive to ignore.
*
*		reg		The register to access. If -1 then this is ignored.
*
*	Returned:
*
*		>= 0	The number of errors encounterred.
*
******************************************************************************/

static int _reg_read(int fd, int line, int index, u32 reg)
{
	int			errs	= 0;
	int			i;
	gsc_reg_t	parm;

	for (;;)	// A convenience loop.
	{
		if ((int) reg == -1)
		{
			// We're supposed to ignore this register.
			errs	= 0;
			break;
		}

		// Read the register.
		parm.reg	= reg;
		parm.value	= 0;;
		parm.mask	= 0;
		i			= ai32ssc1m_ioctl(fd, AI32SSC1M_IOCTL_REG_READ, &parm);

		if (i < 0)
		{
			// There was a problem.
			errs	= 1;
			printf("FAIL <--- (%d. ", line);

			if (index >= 0)
				printf("index %d, ", index);

			printf("AI32SSC1M_IOCTL_REG_READ, status %d)\n", i);
			break;
		}

		errs	= 0;
		break;
	}

	return(errs);
}



/******************************************************************************
*
*	Function:	_reg_show
*
*	Purpose:
*
*		Read a register and display its content.
*
*	Arguments:
*
*		fd		The handle for the device to access.
*
*		line	The line number where the service call is made.
*
*		index	The table index to report, or negactive to ignore.
*
*		reg		The register to access. If -1 then this is ignored.
*
*	Returned:
*
*		>= 0	The number of errors encounterred.
*
******************************************************************************/

static int _reg_show(int fd, int line, int index, u32 reg)
{
	int			errs	= 0;
	int			i;
	gsc_reg_t	parm;

	for (;;)	// A convenience loop.
	{
		if ((int) reg == -1)
		{
			// We're supposed to ignore this register.
			errs	= 0;
			break;
		}

		// Read the register.
		parm.reg	= reg;
		parm.value	= 0;;
		parm.mask	= 0;
		i			= ai32ssc1m_ioctl(fd, AI32SSC1M_IOCTL_REG_READ, &parm);

		if (i < 0)
		{
			// There was a problem.
			errs	= 1;
			printf("FAIL <--- (%d. ", line);

			if (index >= 0)
				printf("index %d, ", index);

			printf("AI32SSC1M_IOCTL_REG_READ, status %d)\n", i);
			break;
		}
		else
		{
			errs	= 0;
			printf("(0x%lX) ", (long) parm.value);
		}

		break;
	}

	fflush(stdout);
	return(errs);
}



/******************************************************************************
*
*	Function:	_reg_test
*
*	Purpose:
*
*		Verify that the specified register bits have the specified values.
*
*	Arguments:
*
*		fd		The handle for the device to access.
*
*		line	The line number where the service call is made.
*
*		index	The table index to report, or negactive to ignore.
*
*		reg		The register to access. If -1 then this is ignored.
*
*		mask	The set of register bits of interest.
*
*		value	The value to apply/check for the bits of interest.
*
*	Returned:
*
*		>= 0	The number of errors encounterred.
*
******************************************************************************/

static int _reg_test(int fd, int line, int index, u32 reg, u32 mask, u32 value)
{
	int			errs	= 0;
	int			i;
	gsc_reg_t	parm;

	for (;;)	// A convenience loop.
	{
		if ((int) reg == -1)
		{
			// We're supposed to ignore this register.
			errs	= 0;
			break;
		}

		if (value & ~mask)
		{
			// Excess value bits are set.
			errs	= 1;
			printf("FAIL <--- (%d. ", line);

			if (index >= 0)
				printf("index %d, ", index);

			printf(	"INTERNAL ERROR, value 0x%lX, mask 0x%lX)\n",
					(long) value,
					(long) mask);
			break;
		}

		// Read the register.
		parm.reg	= reg;
		parm.value	= 0;;
		parm.mask	= 0;
		i			= ai32ssc1m_ioctl(fd, AI32SSC1M_IOCTL_REG_READ, &parm);

		if (i < 0)
		{
			// There was a problem.
			errs	= 1;
			printf("FAIL <--- (%d. ", line);

			if (index >= 0)
				printf("index %d, ", index);

			printf("AI32SSC1M_IOCTL_REG_READ, status %d)\n", i);
			break;
		}

		if ((parm.value & mask) == (value & mask))
		{
			// The register bits of interest are as expected.
			errs	= 0;
			break;
		}

		// The register bits of interest are NOT as expected.
		errs	= 1;
		printf("FAIL <--- (%d. ", line);

		if (index >= 0)
			printf("index %d, ", index);

		printf(	"expected 0x%lX, got 0x%lX (0x%08lX))\n",
				(long) value,
				(long) parm.value & mask,
				(long) parm.value);

		break;
	}

	return(errs);
}



/******************************************************************************
*
*	Function:	_service_perform
*
*	Purpose:
*
*		Perform the specified services on the given structure.
*
*	Arguments:
*
*		fd		The handle for the device to access.
*
*		line	The line number where the service call is made.
*
*		index	A table index to report. Ignore this if negative.
*
*		service	The service to perform.
*
*		item	The structure to process.
*
*	Returned:
*
*		>= 0	The number of errors encounterred.
*
******************************************************************************/

static int _service_perform(
	int						fd,
	int						line,
	int						index,
	service_t				service,
	const service_data_t*	item)
{
	int	errs	= 0;

	switch (service)
	{
		default:
		case SERVICE_NORMAL:
		case SERVICE_END_LIST:

			errs	= 1;
			printf(	"FAIL <---  (%d. INTERNAL ERROR on line %d: ", line, __LINE__);

					if (index >= 0)
						printf("index %d, ", index);

			printf("service %d)\n", (int) service);
			break;

		case SERVICE_IOCTL_GET:

			errs	= _ioctl_get(fd, line, index, item->cmd, item->arg);
			break;

		case SERVICE_IOCTL_SET:

			errs	= _ioctl_set(fd, line, index, item->cmd, item->arg);
			break;

		case SERVICE_REG_MOD:

			errs	= _reg_mod(fd, line, index, item->reg, item->mask, item->value);
			break;

		case SERVICE_REG_READ:

			errs	= _reg_read(fd, line, index, item->reg);
			break;

		case SERVICE_REG_SHOW:

			errs	= _reg_show(fd, line, index, item->reg);
			break;

		case SERVICE_REG_TEST:

			errs	= _reg_test(fd, line, index, item->reg, item->mask, item->value);
			break;

		case SERVICE_SLEEP:

			errs	= 0;
			os_sleep_ms(item->arg * 1000);
			break;
	}

	return(errs);
}



/******************************************************************************
*
*	Function:	_service_item
*
*	Purpose:
*
*		Perform the specified services on the given structure.
*
*	Arguments:
*
*		fd		The handle for the device to access.
*
*		line	The line number where the service call is made.
*
*		index	A table index to report. Ignore this if negative.
*
*		item	The structure to process.
*
*		srv1	The first primary service to perform.
*
*		srv2	The second primary service to perform.
*
*	Returned:
*
*		>= 0	The number of errors encounterred.
*
******************************************************************************/

static int _service_item(
	int						fd,
	int						line,
	int						index,
	const service_data_t*	item,
	service_t				srv1,
	service_t				srv2)
{
	int	errs	= 0;

	if (item->service == SERVICE_NORMAL)
	{
		errs	= _service_perform(fd, line, index, srv1, item);

		if (errs == 0)
			errs	= _service_perform(fd, line, index, srv2, item);
	}
	else
	{
		errs	= _service_perform(fd, line, index, item->service, item);
	}

	return(errs);
}



/******************************************************************************
*
*	Function:	_service_list
*
*	Purpose:
*
*		Go through the list and, for each element, perform the specified
*		services.
*
*	Arguments:
*
*		fd		The handle for the device to access.
*
*		line	The line number where the service call is made.
*
*		index	A table index to report. Ignore this if negative.
*
*		list	The list to go through.
*
*	Returned:
*
*		>= 0	The number of errors encounterred.
*
******************************************************************************/

static int _service_list(
	int						fd,
	int						line,
	const service_data_t*	list,
	service_t				srv1,
	service_t				srv2)
{
	int	errs	= 0;
	int	i;

	for (i = 0; (errs == 0) && (list[i].service != SERVICE_END_LIST); i++)
		errs	= _service_item(fd, line, i, &list[i], srv1, srv2);

	return(errs);
}



/******************************************************************************
*
*	Function:	service_ioctl_reg_get_list
*
*	Purpose:
*
*		Go through the list and, for each element, perform the specified
*		services.
*
*	Arguments:
*
*		fd		The handle for the device to access.
*
*		line	The line number where the service call is made.
*
*		list	The list to go through.
*
*	Returned:
*
*		>= 0	The number of errors encounterred.
*
******************************************************************************/

int service_ioctl_reg_get_list(int fd, int line, const service_data_t* list)
{
	int	errs;

	errs	= _service_list(fd, line, list, SERVICE_REG_MOD, SERVICE_IOCTL_GET);
	return(errs);
}



/******************************************************************************
*
*	Function:	service_ioctl_set_get_list
*
*	Purpose:
*
*		Go through the list and, for each element, perform the specified
*		services.
*
*	Arguments:
*
*		fd		The handle for the device to access.
*
*		line	The line number where the service call is made.
*
*		list	The list to go through.
*
*	Returned:
*
*		>= 0	The number of errors encounterred.
*
******************************************************************************/

int service_ioctl_set_get_list(int fd, int line, const service_data_t* list)
{
	int	errs;

	errs	= _service_list(fd, line, list, SERVICE_IOCTL_SET, SERVICE_IOCTL_GET);
	return(errs);
}



/******************************************************************************
*
*	Function:	service_ioctl_set_reg_list
*
*	Purpose:
*
*		Go through the list and, for each element, perform the specified
*		services.
*
*	Arguments:
*
*		fd		The handle for the device to access.
*
*		line	The line number where the service call is made.
*
*		list	The list to go through.
*
*	Returned:
*
*		>= 0	The number of errors encounterred.
*
******************************************************************************/

int service_ioctl_set_reg_list(int fd, int line, const service_data_t* list)
{
	int	errs;

	errs	= _service_list(fd, line, list, SERVICE_IOCTL_SET, SERVICE_REG_TEST);
	return(errs);
}



/******************************************************************************
*
*	Function:	service_test_set_reg_reg_get
*
*	Purpose:
*
*		Generate a service_data_t list for processing, then submit it for
*		testing. The processing performed is to request the IOCTL service and
*		verify the register results, then modify the register and verify the
*		IOCTL results.
*
*	Arguments:
*
*		fd		The handle for the device to access.
*
*		line	The line number where the service call is made.
*
*		cmd		This is the IOCTL command option to test.
*
*		args	This is the list of supported IOCTL command options. The list
*				is terminated with a -1.
*
*		reg		This is the register that the option affects.
*
*		begin	This is the high field bit index that the option affects.
*
*		end		This is the low field bit index that the option affects.
*
*	Returned:
*
*		>= 0	The number of errors encounterred.
*
******************************************************************************/

int service_test_set_reg_reg_get(int fd, int line, s32 cmd, const s32* args, s32 reg, int begin, int end)
{
	static service_data_t	list[100];

	int	errs	= 0;
	int	i;
	u32	mask	= GSC_FIELD_ENCODE(0xFFFFFFFF, begin, end);

	memset(list, 0, sizeof(list));

	for (i = 0;; i++)
	{
		if (i >= SIZEOF_ARRAY(list))
		{
			errs++;
			printf("FAIL <---  (%d. INTERNAL ERROR on line %d: i %d)\n", line, __LINE__, i);
			list[i - 1].service	= SERVICE_END_LIST;
			break;
		}

		if (args[i] < 0)
		{
			list[i].service	= SERVICE_END_LIST;
			break;
		}

		list[i].service	= SERVICE_NORMAL;
		list[i].cmd		= cmd;
		list[i].arg		= args[i];
		list[i].reg		= reg;
		list[i].mask	= mask;
		list[i].value	= args[i] << end;
	}

	errs	+= service_ioctl_set_reg_list(fd, line, list);
	errs	+= service_ioctl_reg_get_list(fd, line, list);

	errs	+= service_ioctl_set_reg_list(fd, line, list);
	errs	+= service_ioctl_reg_get_list(fd, line, list);

	return(errs);
}



/******************************************************************************
*
*	Function:	service_test_set_get
*
*	Purpose:
*
*		Generate a service_data_t list for processing, then submit it for
*		testing. The processing performed is to request the IOCTL service and
*		verify that the IOCTL returns the same setting.
*
*	Arguments:
*
*		fd		The handle for the device to access.
*
*		line	The line number where the service call is made.
*
*		cmd		This is the IOCTL command option to test.
*
*		args	This is the list of supported IOCTL command options. The list
*				is terminated with a -1.
*
*	Returned:
*
*		>= 0	The number of errors encounterred.
*
******************************************************************************/

int service_test_set_get(int fd, int line, s32 cmd, const s32* args)
{
	static service_data_t	list[100];

	int	errs	= 0;
	int	i;

	memset(list, 0, sizeof(list));

	for (i = 0;; i++)
	{
		if (i >= SIZEOF_ARRAY(list))
		{
			errs++;
			printf("FAIL <---  (%d. INTERNAL ERROR on line %d: i %d)\n", line, __LINE__, i);
			list[i - 1].service	= SERVICE_END_LIST;
			break;
		}

		if (args[i] < 0)
		{
			list[i].service	= SERVICE_END_LIST;
			break;
		}

		list[i].service	= SERVICE_NORMAL;
		list[i].cmd		= cmd;
		list[i].arg		= args[i];
		list[i].reg		= -1;
		list[i].mask	= 0;
		list[i].value	= 0;
	}

	errs	+= service_ioctl_set_get_list(fd, line, list);

	errs	+= service_ioctl_set_get_list(fd, line, list);

	return(errs);
}


